import io
import json
import time
from typing import Generic, Type, TypeVar

import httpx
import polars as pl
from bayesline.api import (
    AsyncTask,
    AsyncTasksApi,
    Task,
    TaskError,
    TaskProgress,
    TasksApi,
)
from pydantic import BaseModel
from tqdm import tqdm

from bayesline.apiclient import ApiClient, AsyncApiClient

T = TypeVar("T")
P = TypeVar("P", bound=BaseModel)


def _raise_for_status(response: httpx.Response, exc_types: list[Type]) -> None:
    if response.status_code in {200, 202}:
        return

    def _unescape(s: str) -> str:
        try:
            return json.loads(s)
        except json.JSONDecodeError:
            return s

    details = response.json()["detail"]
    stacktrace = details.get("stacktrace", "")
    task_error = TaskError(stacktrace)

    if "type" in details:
        if details["type"] == "KeyError":
            raise KeyError(_unescape(details["message"])) from task_error
        elif details["type"] == "ValueError":
            raise ValueError(_unescape(details["message"])) from task_error

        for exc_type in exc_types:
            if exc_type.__name__ == details["type"]:
                raise exc_type(_unescape(details["message"])) from task_error

    raise task_error


class TaskClient(Task[T], Generic[T]):

    def __init__(
        self,
        client: ApiClient,
        task_id: str,
        tqdm_progress: bool = False,
        exc_types: list[Type] | None = None,
    ) -> None:
        self._api_client = client
        self._task_id = task_id
        self._tqdm_progress = tqdm_progress
        self._exc_types = exc_types or []

    def raise_for_status(self, response: httpx.Response) -> None:
        _raise_for_status(response, self._exc_types)

    @property
    def task_id(self) -> str:
        return self._task_id

    def get_progress(self) -> TaskProgress:
        response = self._api_client.get(f"{self._task_id}/progress")
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        return TaskProgress.model_validate(response.json())

    def is_ready(self) -> bool:
        response = self._api_client.get(f"{self._task_id}/ready")
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        return response.json()

    def wait_result(self, timeout: float = -1.0, check_interval: float = 1.0) -> T:
        start_time = time.time()

        pbar = None
        if self._tqdm_progress:
            pbar = tqdm(total=100)

        try:
            self._update_progress(self._tqdm_progress, pbar)
            while not self.is_ready():
                time.sleep(check_interval)
                self._update_progress(self._tqdm_progress, pbar)
                if 0 < timeout < time.time() - start_time:
                    raise TaskError(
                        f"Task {self.task_id} time out after {timeout} seconds."
                    )
            if pbar is not None:
                pbar.update(80 - pbar.n)
                pbar.set_postfix_str("Transferring data")
            return self.get_result()
        finally:
            if pbar is not None:
                pbar.update(100 - pbar.n)
                pbar.set_postfix_str("Done")
                pbar.close()

    def wait_ready(self, timeout: float = -1.0, check_interval: float = 1.0) -> None:
        start_time = time.time()

        pbar = None
        if self._tqdm_progress:
            pbar = tqdm(total=100)

        try:
            self._update_progress(self._tqdm_progress, pbar)
            while not self.is_ready():
                time.sleep(check_interval)
                self._update_progress(self._tqdm_progress, pbar)
                if 0 < timeout < time.time() - start_time:
                    raise TaskError(
                        f"Task {self.task_id} time out after {timeout} seconds."
                    )
            if pbar is not None:
                pbar.update(80 - pbar.n)
                pbar.set_postfix_str("Transferring data")
        finally:
            if pbar is not None:
                pbar.update(100 - pbar.n)
                pbar.set_postfix_str("Done")
                pbar.close()

    def _update_progress(self, progress: bool, pbar: tqdm | None) -> None:
        if progress:
            task_progress = self.get_progress()
            if pbar is not None:
                pbar.update(int(task_progress.last_progress * 0.75 - pbar.n))
                postfix = task_progress.last_message.ljust(25)
                postfix += f" - {task_progress.last_context}"
                pbar.set_postfix_str(postfix)


class AsyncTaskClient(AsyncTask[T], Generic[T]):

    def __init__(
        self,
        client: AsyncApiClient,
        task_id: str,
        tqdm_progress: bool = False,
        exc_types: list[Type] | None = None,
    ) -> None:
        self._api_client = client
        self._task_id = task_id
        self._tqdm_progress = tqdm_progress
        self._exc_types = exc_types or []

    def raise_for_status(self, response: httpx.Response) -> None:
        _raise_for_status(response, self._exc_types)

    @property
    def task_id(self) -> str:
        return self._task_id

    async def get_progress(self) -> TaskProgress:
        response = await self._api_client.get(f"{self._task_id}/progress")
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        return TaskProgress.model_validate(response.json())

    async def is_ready(self) -> bool:
        response = await self._api_client.get(f"{self._task_id}/ready")
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        return response.json()

    async def wait_result(
        self, timeout: float = -1.0, check_interval: float = 1.0
    ) -> T:
        start_time = time.time()

        pbar = None
        if self._tqdm_progress:
            pbar = tqdm(total=100)

        try:
            await self._update_progress(self._tqdm_progress, pbar)
            while not await self.is_ready():
                time.sleep(check_interval)
                await self._update_progress(self._tqdm_progress, pbar)
                if 0 < timeout < time.time() - start_time:
                    raise TaskError(
                        f"Task {self.task_id} time out after {timeout} seconds."
                    )
            if pbar is not None:
                pbar.update(80 - pbar.n)
                pbar.set_postfix_str("Transferring data")
            return await self.get_result()
        finally:
            if pbar is not None:
                pbar.update(100 - pbar.n)
                pbar.set_postfix_str("Done")
                pbar.close()

    async def wait_ready(
        self, timeout: float = -1.0, check_interval: float = 1.0
    ) -> None:
        start_time = time.time()

        pbar = None
        if self._tqdm_progress:
            pbar = tqdm(total=100)

        try:
            await self._update_progress(self._tqdm_progress, pbar)
            while not await self.is_ready():
                time.sleep(check_interval)
                await self._update_progress(self._tqdm_progress, pbar)
                if 0 < timeout < time.time() - start_time:
                    raise TaskError(
                        f"Task {self.task_id} time out after {timeout} seconds."
                    )
            if pbar is not None:
                pbar.update(80 - pbar.n)
                pbar.set_postfix_str("Transferring data")
        finally:
            if pbar is not None:
                pbar.update(100 - pbar.n)
                pbar.set_postfix_str("Done")
                pbar.close()

    async def _update_progress(self, progress: bool, pbar: tqdm | None) -> None:
        if progress:
            task_progress = await self.get_progress()
            if pbar is not None:
                pbar.update(int(task_progress.last_progress * 0.75 - pbar.n))
                postfix = task_progress.last_message.ljust(25)
                postfix += f" - {task_progress.last_context}"
                pbar.set_postfix_str(postfix)


class DataFrameTaskClient(TaskClient[pl.DataFrame]):

    def get_result(self) -> pl.DataFrame:
        response = self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )

        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = []
            for path in response.json():
                response = self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                dfs.append(pl.read_parquet(io.BytesIO(response.content)))
            return pl.concat(dfs)
        else:
            self.raise_for_status(response)
            return pl.read_parquet(io.BytesIO(response.content))


class DataFrameDictTaskClient(TaskClient[dict[str, pl.DataFrame]]):

    def get_result(self) -> dict[str, pl.DataFrame]:
        response = self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )

        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = {}
            for path in response.json():
                response = self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                name = response.headers["Content-Disposition"].split("filename=")[1]
                name = name.split(".")[0]
                dfs[name] = pl.read_parquet(io.BytesIO(response.content))
            return dfs
        elif response.status_code == 204:
            return {}
        else:
            self.raise_for_status(response)
            name = response.headers["Content-Disposition"].split("filename=")[1]
            name = name.split(".")[0]
            return {name: pl.read_parquet(io.BytesIO(response.content))}


class LazyFrameTaskClient(TaskClient[pl.LazyFrame]):

    def get_result(self) -> pl.LazyFrame:
        response = self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )

        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = []
            for path in response.json():
                response = self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                dfs.append(pl.read_parquet(io.BytesIO(response.content)).lazy())
            return pl.concat(dfs)
        else:
            self.raise_for_status(response)
            return pl.read_parquet(io.BytesIO(response.content)).lazy()


class PydanticTaskClient(TaskClient[P], Generic[P]):

    def __init__(
        self,
        model_class: type[P],
        client: ApiClient,
        task_id: str,
        tqdm_progress: bool = False,
        exc_types: list[Type] | None = None,
    ) -> None:
        super().__init__(client, task_id, tqdm_progress, exc_types)
        self._model_class = model_class

    def get_result(self) -> P:
        response = self._api_client.get(
            f"{self._task_id}/result", params={"type": "json"}
        )
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        self.raise_for_status(response)
        return self._model_class.model_validate(response.json())


class AsyncDataFrameTaskClient(AsyncTaskClient[pl.DataFrame]):

    async def get_result(self) -> pl.DataFrame:
        response = await self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = []
            for path in response.json():
                response = await self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                dfs.append(pl.read_parquet(io.BytesIO(response.content)))
            return pl.concat(dfs)
        else:
            self.raise_for_status(response)
            return pl.read_parquet(io.BytesIO(response.content))


class AsyncDataFrameDictTaskClient(AsyncTaskClient[dict[str, pl.DataFrame]]):

    async def get_result(self) -> dict[str, pl.DataFrame]:
        response = await self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )

        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = {}
            for path in response.json():
                response = await self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                name = response.headers["Content-Disposition"].split("filename=")[1]
                name = name.split(".")[0]
                dfs[name] = pl.read_parquet(io.BytesIO(response.content))
            return dfs
        elif response.status_code == 204:
            return {}
        else:
            self.raise_for_status(response)
            name = response.headers["Content-Disposition"].split("filename=")[1]
            name = name.split(".")[0]
            return {name: pl.read_parquet(io.BytesIO(response.content))}


class AsyncLazyFrameTaskClient(AsyncTaskClient[pl.LazyFrame]):

    async def get_result(self) -> pl.LazyFrame:
        response = await self._api_client.get(
            f"{self._task_id}/result", params={"type": "dataframe"}
        )
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        elif response.status_code == 206:
            dfs = []
            for path in response.json():
                response = await self._api_client.get(
                    f"{self._task_id}/result",
                    params={"type": "dataframe", "path": path},
                )
                dfs.append(pl.read_parquet(io.BytesIO(response.content)).lazy())
            return pl.concat(dfs).lazy()
        else:
            self.raise_for_status(response)
            return pl.read_parquet(io.BytesIO(response.content)).lazy()


class AsyncPydanticTaskClient(AsyncTaskClient[P], Generic[P]):

    def __init__(
        self,
        model_class: type[P],
        client: AsyncApiClient,
        task_id: str,
        tqdm_progress: bool = False,
        exc_types: list[Type] | None = None,
    ) -> None:
        super().__init__(client, task_id, tqdm_progress, exc_types)
        self._model_class = model_class

    async def get_result(self) -> P:
        response = await self._api_client.get(
            f"{self._task_id}/result", params={"type": "json"}
        )
        if response.status_code == 404:
            raise KeyError(f"Task {self.task_id} not found")
        self.raise_for_status(response)
        return self._model_class.model_validate(response.json())


class TasksClient(TasksApi):

    def __init__(self, client: ApiClient) -> None:
        self.api_client = client.append_base_path("tasks")

    def get_task_progress(self, task_id: str) -> TaskProgress:
        response = self.api_client.get(f"{task_id}/progress")
        if response.status_code == 404:
            raise KeyError(f"Task {task_id} not found")
        return TaskProgress.model_validate(response.json())


class AsyncTasksClient(AsyncTasksApi):

    def __init__(self, client: AsyncApiClient) -> None:
        self.api_client = client.append_base_path("tasks")

    async def get_task_progress(self, task_id: str) -> TaskProgress:
        response = await self.api_client.get(f"{task_id}/progress")
        return TaskProgress.model_validate(response.json())
